#include "alloc.h"

/**********************************************************************************/
/*** Funcoes de alocacao de memoria*/
/**********************************************************************************/

/*Aloca memoria para um vetor de inteiros*/
int* malloc_int(int n)
{
	int *v = (int *) malloc (n*sizeof(int));
	if (v == NULL)
	{
		printf("Memory Error!\n");
		fprintf(Log, "Memory Error!\n");
		exit(1);
	}
	return v;
}

/*Aloca memoria para uma string*/
unsigned char* string_alloc (int size)
{
	unsigned char* s = (unsigned char*) malloc
	(size*sizeof(char));
	if (s == NULL)
	{
		printf ("Not enought memory\n");
		fprintf (Log,"Not enought memory\n");
		exit(1);
	}
	return s;
}

partition** part_list_alloc (int n)
{
	int i;
	partition** part_list = (partition**) malloc (n*sizeof(partition*));
	if (part_list == NULL)
	{
		printf ("Not enought memory \n");
		fprintf (Log,"Not enought memory \n");
		exit(1);
	}
	for (i = 0; i < n; i++)
	{
		part_list[i] = NULL;
	}
	return part_list;
}

/*Aloca memoria e inicializa as informacoes para a particao*/
partition* part_alloc(unsigned int type, LONG high, DWORD low, int disk_id)
{
	partition* x = (partition*) malloc (sizeof(partition));
	if (x == NULL)
	{
		printf("Erro ao alocar memoria para a particao\n");
		fprintf(Log, "Erro ao alocar memoria para a particao\n");
		exit(1);
	}
	x->type = type;
	x->high = high;
	x->low = low;
	x->disk_id = disk_id;
	x->sb = NULL;
	return x;
}

/*Aloca memoria e inicializa as informacoes para o superbloco, se a particao for Ext*/
superblock* superblock_alloc (unsigned char* buffer)
{
	unsigned int adjust;
	superblock* x;
	x = (superblock *) malloc (sizeof (superblock));
	if (x == NULL)
	{
		printf("Erro ao alocar memoria para o superblock\n");
		fprintf(Log, "Erro ao alocar memoria para o superblock\n");
		exit(1);
	}	
	adjust = 0;
	x->type = convert_to_int (extract_from_buffer(buffer, SB_SIG_OFF, SB_SIG_OFF +1), 2);
	if (x->type != EXT_SIG)
	{
		adjust = 512;
		x->type = convert_to_int (extract_from_buffer(buffer, SB_SIG_OFF+adjust, SB_SIG_OFF+adjust +1), 2);
		if (x->type != EXT_SIG)
		{
			return NULL;
		}
	}
	x->total_inodes = convert_to_int (extract_from_buffer(buffer, SB_T_IN_OFF+adjust, SB_T_IN_OFF+adjust +3), 4);
	x->total_blocks = convert_to_int (extract_from_buffer(buffer, SB_T_BKS_OFF+adjust, SB_T_BKS_OFF+adjust +3), 4);
	x->block_size = 1024 << (convert_to_int (extract_from_buffer(buffer, SB_BK_SZ_OFF+adjust, SB_BK_SZ_OFF+adjust +3), 4));
	x->blocks_per_bg = convert_to_int (extract_from_buffer(buffer, SB_BKS_PER_BG_OFF+adjust, SB_BKS_PER_BG_OFF+adjust +3), 4);
	x->inodes_per_bg = convert_to_int (extract_from_buffer(buffer, SB_IN_PER_BG_OFF+adjust, SB_IN_PER_BG_OFF+adjust +3), 4);
	x->version = convert_to_int (extract_from_buffer(buffer, SB_VS_OFF+adjust, SB_VS_OFF+adjust +1), 2);
	x->reserved_blocks = convert_to_int (extract_from_buffer(buffer, SB_RV_BKS_OFF+adjust, SB_RV_BKS_OFF+adjust+3), 4);
	x->gd_size = convert_to_int (extract_from_buffer(buffer, SB_GD_SZ_OFF+adjust, SB_GD_SZ_OFF+adjust+1), 2);
	
	if (convert_to_int (extract_from_buffer(buffer, 0x60+adjust, 0x60+adjust+1), 2) & 0x200)
	{
		x->flex_bg = 1 << buffer[0x174+adjust];
	}
	else
	{
		x->flex_bg = 0;
	}
	if (x->version < 1)
	{
		x->inode_size = 128;
	}
	else
	{
		x->inode_size = convert_to_int (extract_from_buffer(buffer, SB_IN_SZ_OFF+adjust, SB_IN_SZ_OFF+adjust +1), 2);
	}
	x->bgs = amount_of_bg (x->total_blocks, x->blocks_per_bg, x->total_inodes, x->inodes_per_bg);
	
	
	return x;
}


file_entry* file_entry_alloc()
{
	file_entry* x = (file_entry*) malloc (sizeof(file_entry));
	if (x == NULL)
	{
		printf("Memoria insuficiente\n");
		exit(1);
	}
	x->dir = NULL;
	return x;
}

bklist* bklist_alloc()
{
	bklist* x;
	x = (bklist*) malloc (sizeof(bklist));
	if (x == NULL)
	{
		printf("Falta de memoria ao alocar bklist*\n");
		fprintf(Log, "Falta de memoria ao alocar bklist*\n");
		exit(1);
	}
	x->next = NULL;
	return x;
}