#include "calc.h"

/**********************************************************************************/
/*** Facilitadores*/
/**********************************************************************************/

/*  Little Endian -> Big Endian */
unsigned int convert_to_int (unsigned char *b, int size)
{
    unsigned int i = 0;
    if (size == 2)
    {
        i = b[1];
        i = i << 8;
        i = i + b[0];
        return i;
    }
    else if (size == 3)
    {
        i = b[2];
        i = i << 8;
        i = i + b[1];
        i = i << 8;
        i = i + b[0];
        return i;
    }
    else if (size == 4)
    {
        i = b[3];
        i = i << 8;
        i = i + b[2];
        i = i << 8;
        i = i + b[1];
        i = i << 8;
        i = i + b[0];
        return i;
    }
    else
    {
        printf ("Wrong use!!!! Check your code NOW\n");
		fprintf (Log, "Wrong use!!!! Check your code NOW\n");
        exit (1);
    }
}

/*Devolve um pedao sequencial de um buffer*/
unsigned char* extract_from_buffer (unsigned char* buffer, int from, int to)
{
    int i;
    unsigned char *b = string_alloc (to - from + 2);
    for (i = 0; i < (to - from + 1); i++)
        b[i] = buffer[from+i];
	b[i] = '\0';
    return b;
}
/**********************************************************************************/

/*Devolve o tamanho de um setor do disco*/
unsigned int disk_bytes_per_sector(int disk_id)
{
	HANDLE disk;
	int success;
	DWORD count;
	DISK_GEOMETRY_EX geo;
	disk = open_disk(disk_id);
	success = DeviceIoControl(disk, IOCTL_DISK_GET_DRIVE_GEOMETRY,  NULL, 0, &geo, sizeof(geo), &count, NULL);
	close_disk(disk);
	if ((success == 0) || (geo.Geometry.BytesPerSector < MIN_BYTES_PER_SECTOR))
	{
		return MIN_BYTES_PER_SECTOR;
	}
	else{
		return geo.Geometry.BytesPerSector;
	}
}

/*Devolve a quantidade de trilhas por cilindro*/
unsigned int disk_tracks_per_cylinder(int disk_id)
{
	HANDLE disk;
	DWORD count;
	DISK_GEOMETRY_EX geo;
	disk = open_disk (disk_id);
	DeviceIoControl(disk, IOCTL_DISK_GET_DRIVE_GEOMETRY,  NULL, 0, &geo, sizeof(geo), &count, NULL);
	close_disk(disk);
	return geo.Geometry.TracksPerCylinder;
	
}

/*Devolve a quantidade de setores por trilha*/
unsigned int disk_sectors_per_track(int disk_id)
{
	HANDLE disk;
	DWORD count;
	DISK_GEOMETRY_EX geo;
	disk = open_disk (disk_id);
	DeviceIoControl(disk, IOCTL_DISK_GET_DRIVE_GEOMETRY,  NULL, 0, &geo, sizeof(geo), &count, NULL);
	close_disk(disk);
	return geo.Geometry.SectorsPerTrack;

}



/*Espera o usuario apertar Enter - Usada para depuracao*/
void wait_enter()
{
	while(1)
	{
		getchar();
		break;
	}
}

/*Interpreta a informacao passada e diz a qual tipo de particao se refere*/
int partition_type(unsigned int id)
{
	if (id == 0x5)
	{
		printf("Extended-CHS ");
		fprintf(Log, "Extended-CHS ");
	}
	else if (id == 0x7)
	{
		printf ("NTFS ");
		fprintf (Log, "NTFS ");
	}
	else if (id == 0xf)
	{
		printf("Extended-LBA ");
		fprintf(Log, "Extended-LBA ");
	}
	else if (id == 0x82)
	{
		printf ("Linux Swap ");
		fprintf (Log, "Linux Swap ");
	}
	else if (id == 0x83)
	{
		printf ("Linux native FS ");
		fprintf (Log, "Linux native FS ");
	}
	else
	{
		printf("ID: %x ", 0xff & id);
		fprintf(Log, "ID: %x ", 0xff & id);
	}
	return id;
}

/*No sei se vou usar, mas t ai =P - Se for usar, tem q arrumar*/
unsigned int chs_to_bytes(unsigned int chs, unsigned int max_heads_per_cylinder, unsigned int max_sectors_per_track)
{
	unsigned int cylinders, heads, sectors;
	cylinders = ((chs&0xff0000)>>16) + ((chs&0xc000)>>6);
	heads = chs&0xff;
	sectors = (chs&0x3f00)>>8;
	printf("heads: %x - sectors: %x - cylinders: %x\n",heads, sectors, cylinders);
	return ((cylinders*max_heads_per_cylinder) + heads)*max_sectors_per_track + sectors -1;
}




unsigned int find_inode_block_group (unsigned int inode, unsigned int inodes_per_group)
{
	return (inode - 1)/ inodes_per_group;
}

unsigned int find_inode_index (unsigned int inode, unsigned int inodes_per_group)
{
	return (inode - 1) % inodes_per_group;
}

/*Pq isso  igual ao find_inode_block_group ?*/
unsigned int find_inode_block (unsigned int inode, unsigned int inodes_per_group)
{
	return (inode - 1)/ inodes_per_group;
}

unsigned int amount_of_bg(unsigned int total_blocks, unsigned int blocks_per_bg, unsigned int total_inodes, unsigned int inodes_per_bg)
{
	unsigned int bg_per_part, remains, mem;
	bg_per_part = total_blocks/blocks_per_bg;
	remains = total_blocks%blocks_per_bg;
	if (remains != 0)
		bg_per_part++;
	mem = bg_per_part;
	
	bg_per_part = total_inodes/inodes_per_bg;
	remains = total_inodes%inodes_per_bg;
	if (remains != 0)
		bg_per_part++;
		
	if (mem != bg_per_part)
	{
		printf("Something is very wrong! Amount of blockgroups using inodes calculation and blocks calculation differs\n");
		fprintf(Log,"Something is very wrong! Amount of blockgroups using inodes calculation and blocks calculation differs\n");
	}
	return bg_per_part;
}

unsigned int check_mbr_sign(unsigned char* mbr)
{
	if (convert_to_int(extract_from_buffer(mbr,510,511),2) == 0xAA55)
		return 1;
	return 0;
}
