#include "tocatalog.h"

/**********************************************************************************/
/*** Funcoes nao catalogadas*/
/**********************************************************************************/

file_entry** create_directory (bklist* list, unsigned int flagporca, partition* ext, unsigned int *entries)
{
	unsigned int i,j, inode, name_size, dentry_size, file_type, size;
	file_entry** v;
	unsigned char* buffer;
	buffer = directory_buffer(list, ext, &size);
	
	/*Tamanho do vetor supondo pior caso: arquivos cujo nome contem apenas 1 caracter*/
	v = (file_entry**) malloc ((size/9)*sizeof(file_entry*));
	if (v == NULL)
	{
		printf("Memoria insuficiente\n");
		exit(1);
	}
	
	
	
	for (i = 0, j = 0; i < size; i=i+dentry_size, j++)
	{
		
		file_type = 0xffffffff;
		inode = convert_to_int (extract_from_buffer(buffer,i,i+0x4), 4);
		dentry_size = convert_to_int (extract_from_buffer(buffer,i+0x4,i+0x5), 2);
		if (inode == 0)
		{
			v[j] = NULL;
			break;
		}
		v[j] = file_entry_alloc();
		if (flagporca)
		{
			name_size = buffer[i+0x6];
			file_type = buffer[i+0x7];
			/*printf("name_size: %d\n", name_size);
			fprintf(Log, "name_size: %d\n", name_size);*/
		}
		else
		{
			name_size = convert_to_int(extract_from_buffer(buffer,i+0x6,i+0x7),2);
		}
		v[j]->type = file_type;
		v[j]->inode = inode;
		v[j]->size = 8; /*reservado para uso futuro*/
		v[j]->name = extract_from_buffer(buffer, i+0x8, i+0x8+name_size-1);
		/*printf("|%4d | %-50s", j, v[j]->name);
		fprintf(Log, "|%4d | %-50s", j, v[j]->name);*/
		
		/*printf(" - Inode: %x\n", inode);
		fprintf(Log, " - Inode: %x\n", inode);*/
	}
	*entries = j;
	/*Ordena as entradas*/
	qsort(v,j,sizeof(file_entry*), compare);
	return v;
}

blockgroup** create_bd_list(unsigned char* buffer, unsigned int bd_size, unsigned int bgs, unsigned int block_size)
{
	int i;
	blockgroup** x;
	DWORDLONG begin;
	LONG high;
	DWORD low;
	
	x = (blockgroup**) malloc (bgs*sizeof(blockgroup*));
	if (x == NULL)
	{
		printf("Memoria insuficiente\n");
		exit(1);
	}
	
	for (i = 0; i < bgs; i++)
	{
		x[i] = (blockgroup*) malloc (sizeof(blockgroup));
		if (x[i] == NULL)
		{
			printf("Memoria insuficiente\n");
			exit(1);
		}
		if (bd_size == 64)
		{
			begin = convert_to_int(extract_from_buffer (buffer, i*bd_size+0x28, i*bd_size+0x28+4),4);
			begin = (begin << 32);
			/*printf("DWORDLONG size: %d\n", sizeof(DWORDLONG));
			fprintf(Log,"DWORDLONG size: %d\n", sizeof(DWORDLONG));*/
			x[i]->high_inode_table = convert_to_int(extract_from_buffer (buffer, i*bd_size+0x28, i*bd_size+0x28+4),4);
		}
		else
		{
			begin = 0;
			x[i]->high_inode_table = 0;
		}
		begin = begin + convert_to_int(extract_from_buffer (buffer, i*bd_size+0x8, i*bd_size+0x8+4),4);
		x[i]->low_inode_table = convert_to_int(extract_from_buffer (buffer, i*bd_size+0x8, i*bd_size+0x8+4),4);
		set_high_low (&low, &high, block_size, begin);
		x[i]->high_inode_table = high;
		x[i]->low_inode_table = low;
	}
	return x;
}


partition** create_part_list(int n, int *empty_pos)
{
	int i, j;
	partition** part_list;
	unsigned char* buffer;
	unsigned int sector_size;
	part_list = part_list_alloc(100);
	/*Percorre todos os discos e, em cada disco, cria uma estrutura do tipo partition para cada particao*/
	j = 0;
	for (i = 0; i < n; i++)
	{	
		read_partition_table(i,&part_list, empty_pos);
		/*printf("Foram encontradas %d particoes ate agora\n", *empty_pos);
		fprintf(Log,"Foram encontradas %d particoes ate agora\n", *empty_pos);*/
		sector_size = disk_bytes_per_sector(i);
		for (; j < *empty_pos; j++)
		{			
			/*Se a particao for do tipo Linux Native FS*/
			if (part_list[j]->type == 0x83)
			{
				/*Le as informacoes sobre a particao no disco*/

				buffer = read_disk(i, part_list[j]->high, part_list[j]->low, SB_SZ+sector_size, 0, (SB_OFF/sector_size)*sector_size); /*Troquei SB_SZ de high pra low*/
				/*Criar uma estrutura do tipo superblock se a particao for Ext{2,3,4}  */
				buffer = extract_from_buffer(buffer,SB_OFF%sector_size,(SB_OFF%sector_size) + SB_SZ);
				part_list[j]->sb = superblock_alloc (buffer);
			}
		}
	}
	return part_list;
}


void set_high_low_inode (DWORD *low, LONG *high, unsigned int inode, unsigned int inode_size, unsigned int inodes_per_group , blockgroup** bg)
{
	unsigned int group_index, inode_index;
	DWORDLONG location;
	group_index = find_inode_block_group (inode, inodes_per_group);
	inode_index = find_inode_index (inode, inodes_per_group);
	location = bg[group_index]->high_inode_table;
	location = location << 32;
	location += bg[group_index]->low_inode_table;
	location += inode_size*(inode_index);
	*low = location & 0xffffffff;
	*high = (LONG) (((DWORDLONG) (location >> 32)) & 0xffffffff);
}





/*unsigned int go_to_block_begin ();*/

/*recebe o vetor maroto do inode de tamanho 15*intsize*/
bklist* read_inode_blocks(unsigned char* buffer, partition* ext, unsigned int flags)
{
	if (flags & 0x80000)
		return read_extent_tree(buffer, ext);
	return read_block_addressing(buffer, ext);
}

/*Arrumar essa funcao*/
bklist* read_block_addressing(unsigned char* buffer, partition* ext)
{
	unsigned int i, block;
	bklist* x,*head;
	DWORD low;
	LONG high;
	x = bklist_alloc();
	head = x;
	for(i = 0; i < 12; i++)
	{
		block = convert_to_int(extract_from_buffer(buffer,4*i, 4*i + 0x03),4);
		if (block == 0)
			break;
		x->next = bklist_alloc();
		x = x->next;
		x->len = 1;
		set_high_low (&low, &high, ext->sb->block_size, block);
		x->low = low;
		x->high = high;
	}
	
	return head->next;
}

bklist* read_extent_tree(unsigned char* buffer, partition* ext)
{
	unsigned int i, depth, sign, n, len, pos;
	DWORDLONG block;
	DWORD low, bk_lower;
	LONG high, bk_upper;
	bklist* x;
	bklist* head;
	
	x = bklist_alloc();
	head = x;
	
	/*Lendo a header*/
	sign = convert_to_int(extract_from_buffer(buffer,0x00,0x01),2);
	n = convert_to_int(extract_from_buffer(buffer,0x02,0x03),2);
	depth = convert_to_int(extract_from_buffer(buffer,0x06,0x07),2);
	
	/*Verificando a assinatura*/
	if (sign != 0xF30A)
	{
		printf("Tem algo muito errado. A assinatura da header nao bate na funcao read_extent_tree\n");
		fprintf(Log, "Tem algo muito errado. A assinatura da header nao bate na funcao read_extent_tree\n");
		wait_enter();
		exit(1);
	}
	/*Se for folha*/
	if (depth == 0)
	{
		/*printf("Eh folhaaaaaaaaaaaaaaaa!!!\nTem %d entradas!!!\n", n);
		wait_enter();*/
		for (i = 0; i < n; i++)
		{
			
			pos = convert_to_int(extract_from_buffer(buffer,12*(i+1) + 0x00, 12*(i+1) + 0x03),4);
			len = convert_to_int(extract_from_buffer(buffer,12*(i+1) + 0x04, 12*(i+1) + 0x05),2);
			bk_upper = convert_to_int(extract_from_buffer(buffer,12*(i+1) + 0x06, 12*(i+1) + 0x07),2);
			bk_lower = convert_to_int(extract_from_buffer(buffer,12*(i+1) + 0x08, 12*(i+1) + 0x0B),4);
			block = bk_upper << 16;
			block = block << 16;
			block += bk_lower;
			set_high_low (&low, &high, ext->sb->block_size, block);
			x->next = bklist_alloc();
			x = x->next;
			x->high = high;
			x->low = low;
			x->len = len;
			x->order = pos;
		}
	}
	else
	{
		for (i = 0; i < n; i++)
		{
			while(x->next != NULL) x = x->next;
			pos = convert_to_int(extract_from_buffer(buffer,12*(i+1) + 0x00, 12*(i+1) + 0x03),4);
			bk_lower = convert_to_int(extract_from_buffer(buffer,12*(i+1) + 0x04, 12*(i+1) + 0x07),4);
			bk_upper = convert_to_int(extract_from_buffer(buffer,12*(i+1) + 0x08, 12*(i+1) + 0x09),2);
			block = bk_upper << 16;
			block = block << 16;
			block += bk_lower;
			set_high_low (&low, &high, ext->sb->block_size, block);
			x->next = read_extent_tree(read_disk(ext->disk_id, ext->high, ext->low, ext->sb->block_size, high, low), ext);
			
		}
	}
	
	/*Remove a cabeca*/
	return head->next;
}



unsigned char* directory_buffer(bklist* list, partition* ext, unsigned int *size)
{
	unsigned char* buffer, *aux;
	bklist* x;
	unsigned int i, lencontrol;
	DWORD low;
	LONG high;
	*size = 0;
	x = list;
	while (x != NULL)
	{
		*size += x->len * ext->sb->block_size;
		x = x->next;
	}
	x = list;
	
	lencontrol = 0;
	buffer = string_alloc(*size);
	for (i = 0; i < *size; i++)
	{
		if (i % ext->sb->block_size == 0)
		{
			if (lencontrol == 0)
			{
				high = x->high;
				low = x->low;
			}
			else
			{
				low += ext->sb->block_size;
				/*se deu overflow*/
				if ((low < x->low) && high == x->high)
				{
					high++;
				}
			}
			
			aux = read_disk(ext->disk_id, ext->high, ext->low, ext->sb->block_size, high, low);
			lencontrol++;
			if(lencontrol == x->len)
			{
				lencontrol = 0;
				x = x->next;
			}
		}
		buffer[i] = aux[i%ext->sb->block_size];
	}
	return buffer;
}

int entrada(file_entry* x)
{
	int i;
	unsigned char* buffer;
	unsigned char** v;
	char c;
	buffer = string_alloc(4096);
	i = 0;
	c = getchar();
	while(c != '\n')
	{
		buffer[i++] = c;
		c = getchar();
	}
	buffer[i] = '\0';
	printf("\n");
	v = parser(buffer);
	/*for (i = 0; v[i] != NULL; i++)
		printf("%s;",v[i]);
	printf("\n");*/
	if (!strcmp((char*) v[0], "cd"))
	{
		if (v[1] != NULL)
		{
			/*Se o segundo parametro for "-n"*/
			if  (!strcmp((char*) v[1], "-n"))
			{
				if (v[2] != NULL)
				{
					return atoi((char*) v[2]);
				}
				else
					return -1;
			}
			else
			{
				for (i = 0; x->dir[i] != NULL; i++)
				{
					/*printf("%s %s\n",(char*) v[1], x->dir[i]->name);*/
					if(!strcmp((char*) v[1], (char *) x->dir[i]->name))
					{
						if(x->dir[i]->type == 0x2)
							return i;
						else
							return -1;
					}
				}
			}
		}
	}
	else if(!strcmp((char*) v[0], "extract"))
	{
		if (v[1] != NULL)
		{
			/*Se o segundo parametro for "-n"*/
			if  (!strcmp((char*) v[1], "-n"))
			{
				if (v[2] != NULL)
				{
					return atoi((char*) v[2]);
				}
				else
					return -1;
			}
			
			else{
				for (i = 0; x->dir[i] != NULL; i++)
				{
					/*printf("%s %s\n",(char*) v[1], x->dir[i]->name);*/
					if(!strcmp((char*) v[1], (char *) x->dir[i]->name))
					{
						if(x->dir[i]->type == 0x1)
							return i;
						else
							return -1;
					}
				}
			}
		}
	}
	else if(!strcmp((char*) v[0], "help"))
	{
		help();
		return -2;
	}
	else if (!strcmp((char*) v[0], "list"))
	{
		list(x->dir);
		return -2;
	}
	else if (!strcmp((char*) v[0], "cdest"))
	{
		if (v[1] != NULL)
		{
			SetCurrentDirectory((LPSTR) v[1]);
			return -2; /*Significa que o comando foi bem sucedido, mas nao mudou a pasta ext*/
		}
	}
	else if (!strcmp((char*) v[0], "exit"))
	{
		exit(1);
	}
	else if(!strcmp((char*) v[0], "cpart"))
	{
		if (v[1] != NULL)
		{
			/*Se o segundo parametro for "-n"*/
			if  (!strcmp((char*) v[1], "-n"))
			{
				if (v[2] != NULL)
				{
					return 0x1000000 + atoi((char*) v[2]);
				}
				else
					return -1;
			}
		}
	}
	return -1;
}


unsigned char** parser(unsigned char* mander)
{
	int flag, i, j, k;
	unsigned char** v;
	v = (unsigned char**) malloc (16*sizeof(unsigned char*));
	if (v == NULL){
		printf("Erro ao alocar memoria\n");
		exit(1);
	}
	for (i = 0; i < 16; i++)
	{
		v[i] = NULL;
	}
	/*Pega os parametros*/
	j = 0;
	k = 0;
	flag = 0; 
	v[k] = string_alloc(256);
	for (i = 0; mander[i] != '\0'; i++)
	{
		if (validChar(mander[i]))
		{
			v[k][j++] = mander[i];
			flag = 1;
		}
		else if ((mander[i] == '\\') && (mander[i+1] == ' '))
		{
		    v[k][j++] = ' ';
		    i++;
		}
		else if (flag == 1)
		{
			v[k][j] = '\0';
			k++;
			j = 0;
			v[k] = string_alloc(256);
			flag = 0;
			while (mander[i] == ' ') i++;
			if (mander[i] == ',') i++;
			i--;
		}
	}
	v[k][j] = '\0';
	return v;
}

int validChar (char c)
{
	if (c >= 'A' && c <= 'Z')
		return 1;
	else if (c >= 'a' && c <= 'z')
		return 1;
	else if (c >= '0' && c <= '9')
		return 1;
	else if (c == '.')
		return 1;
	else if (c == '-')
		return 1;
	else if (c == '_')
		return 1;
	else if (c == '@')
		return 1;
	else if (c == ':')
		return 1;
	return 0;
}

void extract(unsigned char* name,unsigned int inode, bklist* x, partition* ext)
{
	DWORDLONG i, size;
	unsigned char* buffer;
	unsigned int lencontrol, percents, over;
	LONG high;
	DWORD low;
	FILE* arquivo;
	
	get_file_size_inode(inode, ext,&size);
	lencontrol = 0;
	over = size/100;
	arquivo = fopen((char*) name,"wb");
	percents = 0;
	/*printf("%d",x->order);*/
	printf(" Extraindo %s\n", name);
	printf("    Progresso:\n");
	printf("            0%%");
	if (arquivo != NULL)
	{
		for(i = 0; i < size; i++)
		{
			if(i % ext->sb->block_size == 0)
			{
				if (lencontrol == 0)
				{
					high = x->high;
					low = x->low;
				}
				else
				{
					low += ext->sb->block_size;
					/*se deu overflow*/
					if ((low < x->low) && high == x->high)
					{
						high++;
					}
				}
				buffer = read_disk(ext->disk_id, ext->high, ext->low, ext->sb->block_size, high, low);
				lencontrol++;
				if(lencontrol == x->len)
				{
					lencontrol = 0;
					x = x->next;
					/*if (x != NULL)
						printf("->%d",x->order);*/
				} 
			}
			fputc(buffer[i%ext->sb->block_size], arquivo);
			
			/*Imprime barrinha de porcentagem*/
			if ((i/over) >= percents)
			{
				printf ("\r          %3d%%", percents);
				percents++;
			}
		}
	}
	fclose(arquivo);
	printf ("\r          100%%");
	printf("\n %s extraido com sucesso\n", name);
}


void get_file_size_inode(unsigned int inode, partition* ext, DWORDLONG *size)
{
	DWORD low;
	LONG high;
	unsigned char* buffer;
	unsigned int upper, lower;
	unsigned int sector_size = disk_bytes_per_sector(ext->disk_id);
	set_high_low_inode(&low, &high, inode, ext->sb->inode_size, ext->sb->inodes_per_bg , ext->bg);
	
	/*Essa gambiarra eh pq o read_disk soh le multiplos de setor*/
	buffer = read_disk(ext->disk_id, ext->high, ext->low, ext->sb->block_size, high, (low/sector_size)*sector_size);
	buffer = extract_from_buffer(buffer,low%sector_size,(low%sector_size) + ext->sb->inode_size);
	lower = convert_to_int(extract_from_buffer(buffer,0x4,0x7),4);
	upper = convert_to_int(extract_from_buffer(buffer,0x6C,0x6F),4);
	*size = upper < 32;
	*size += lower;
}


void help()
{
	printf("\nExtrator de arquivos do Ext para o Windows:\n");
	printf("  list    - lista o conteudo dos diretorios\n");
	printf("            uso: list\n\n");
	printf("  cd      - altera o diretorio atual da particao Ext\n");
	printf("            uso: cd <nome_do_diretorio>\n");
	printf("                 cd -n <numero_do_diretorio>\n\n");
	printf("  cdest   - altera o diretorio de destino na particao do Windows\n");
	printf("            uso: cdest <nome_do_diretorio>\n\n");
	printf("  cpart   - altera a particao atual\n");
	printf("            uso: cpart -n <numero_da_particao>\n\n");
	printf("  exit    - sai do programa\n");
	printf("            uso: exit\n\n");
	printf("  extract - extrai o arquivo para o diretorio de destino\n");
	printf("            uso: extract <nome_do_arquivo>\n");
	printf("            uso: extract -n <numero_do_arquivo>\n\n");
	printf("  help    - exibe essa mensagem de ajuda\n");
	printf("            uso: help\n\n");
	
}

int compare(const void* xa, const void* xb)
{
	file_entry *a, *b;
	a = *(file_entry**) xa;
	b = *(file_entry**) xb;
	/*Ordenar tipos por ordem decrescente*/
	if(a->type > b->type)
	{
		return -1;
	}
	else if (a->type < b->type)
	{
		return 1;
	}
	/*Nomes por ordem crescente*/
	if (strcmp((char*) a->name,(char *)b->name) > 0)
	{
		return 1;
	}
	return -1;
}

void list(file_entry** v)
{
	int i;
	
	for (i = 0; v[i] != NULL; i++)
	{
		printf("|%4d | %-50s", i, v[i]->name);
		print_file_type(v[i]->type);
		printf("\n");
	}
}

void print_file_type (unsigned int file_type)
{
	if (file_type != 0xffffffff)
	{
		if (file_type == 0x0)
		{
			printf(" | Unknown");
			fprintf(Log, " | Unknown");
		}
		else if (file_type == 0x1)
		{
			printf(" | Regurar File");
			fprintf(Log, " | Regular File");
		}
		else if (file_type == 0x2)
		{
			printf(" | Directory");
			fprintf(Log, " | Directory");
		}
		else if (file_type == 0x3)
		{
			printf(" | Character device file");
			fprintf(Log, " | Character device file");
		}
		else if (file_type == 0x4)
		{
			printf(" | Block device file");
			fprintf(Log, " | Block device file");
		}
		else if (file_type == 0x5)
		{
			printf(" | FIFO");
			fprintf(Log, " | FIFO");
		}
		else if (file_type == 0x6)
		{
			printf(" | Socket");
			fprintf(Log, " | Socket");
		}
		else if (file_type == 0x7)
		{
			printf(" | Symbolic Link");
			fprintf(Log, " | Symbolic Link");
		}
	}
}